(*
	Have an array result of size (nb_1_total+1) x (nb_2_total+1).
	result.(i).(j) contains the esperance and the optimum strategy (continue or stop)
	for someone who already drawn i "1" and j"2".

	Actually every lines are identical (ie the number of "1" drawn isn't relevant):
	for i and j given, one has a probability of loosing everything, another to doule his gain.
	Only the ratio counts: one continue iff he has more chances to double than to loose.

	So one needs only an array of size (nb_2_total+1).
	This problem could have been made harder by increasing the limit on nb_2_total...
*)

type node = {esperance: float; continue:bool}
let dummy = {esperance = 1.; continue = false}

let scanf format = Scanf.bscanf Scanf.Scanning.stdib format;;

for nb_tests=1 to scanf "%i\n" (fun i -> i) do
 let nb_cases, nb_1_total, nb_2_total = scanf "%i %i %i\n" (fun a b c -> a,b,c) in
 let result = Array.make_matrix (succ nb_1_total) (succ nb_2_total) dummy in
	let max_esperance = 2. ** (float nb_2_total) in
	for nb_de_1 = nb_1_total downto 0 do
		result.(nb_de_1).(nb_2_total) <- {esperance = max_esperance; continue=false};
	done;
	for nb_de_2 = pred nb_2_total downto 0 do
		let prudent = 2. ** float nb_de_2 in
		let optimist =
			let hidden_cases = float (nb_cases - nb_1_total - nb_de_2) in
			result.(nb_1_total).(nb_de_2 +1).esperance   *.   float (nb_2_total - nb_de_2) /. hidden_cases (* if we draw "2" *)
				(* impossible to draw "1", everything lost if we draw "0" *)
		in
		result.(nb_1_total).(nb_de_2) <- {esperance = max optimist prudent; continue = optimist > prudent};
	done;
	for nb_de_2 = pred nb_2_total downto 0 do
		for nb_de_1 = pred nb_1_total downto 0 do
			let prudent = 2. ** float nb_de_2 in
			let optimist =
				let hidden_cases = float (nb_cases - nb_de_1 - nb_de_2) in
				result.(nb_de_1).(nb_de_2 +1).esperance
				*. float (nb_2_total - nb_de_2) /. hidden_cases (* si on tire un 2 *)
				+. result.(nb_de_1 +1).(nb_de_2).esperance
				*. float (nb_1_total - nb_de_1) /. hidden_cases (* si on tire un 1 *) in
			result.(nb_de_1).(nb_de_2) <- {esperance = max optimist prudent; continue = optimist > prudent};
		done;
	done;
	Printf.printf "%.2f\n%!" result.(0).(0).esperance;
done;;
(* (\* print the array: *\)
	Printf.printf "\n%!";
	for i=0 to nb_1_total do
  for j=0 to nb_2_total do
  let x = result.(i).(j) in
  Printf.printf " %10.2f %s |" x.esperance (if x.continue then "C" else "S")
  done;
  Printf.printf "\n%!"
	done;;
*)


