(* Uncomment if you want to test interactively in the toplevel. *)
(* #load "nums.cma";; *)
(* #load "camlp4o.cma";; *)

(********* parsing *********)
(* OK, using OCaml Genlex and streams is a bit cheating.
In C, one has to replace products by their value first, then add the terms of th sum. *)

open Genlex;;
let print_token = function Int i -> print_int i | Kwd s -> Printf.printf " (%s) " s | _ -> invalid_arg "print_token";;

type expr =
  |Binop of string * expr * expr
	|Entier of int;;

let lexer = Genlex.make_lexer ["+"; "-"; "*"; "/"];;

let read_operator operators = parser
		| [< 'Kwd op when List.mem op operators >] -> op ;
(* 		| [< >] -> Printf.printf "fail: != %s\n%!" (List.hd operators); raise Stream.Failure *)
;;
let read_binop read_sub_element operators =
	let rec read_test expr1 = parser
		| [< op = read_operator operators;
				 expr2 = read_sub_element;
				 expr = read_test (Binop (op,expr1,expr2)) >] -> expr
		| [< >] -> expr1 in
	parser
		[< expr1 = read_sub_element; expr = read_test expr1 >] -> expr;;

let read_int = parser
  | [< 'Int i >] -> Entier i;;
let read_multiplicative_op = read_binop read_int               ["*"; "/"];;
let read_additive_op       = read_binop read_multiplicative_op ["+"; "-"];;

(********* evaluate something of type expr *********)
let rec eval = function
		Binop ("+",e1,e2) -> eval e1 +. eval e2
	| Binop ("-",e1,e2) -> eval e1 -. eval e2
	| Binop ("*",e1,e2) -> eval e1 *. eval e2
	| Binop ("/",e1,e2) -> eval e1 /. eval e2
	| Binop (_,e1,e2) -> invalid_arg "eval"
	| Entier n -> float n;;
let compare_evaluations goal evaluation =
	if abs_float (float goal -. evaluation) < 0.000_000_000_000_1 then 0
		(* Works for epsilon between 0.01 and 0.000_000_000_000_1, but checked only for the two used inputs. *)
	else compare (float goal) evaluation;;

(* Uncomment this block to use arbitrary precision rationals instead of floats. *)
(*
open Num;;
let rec eval = function
		Binop ("+",e1,e2) -> eval e1 +/ eval e2
	| Binop ("-",e1,e2) -> eval e1 -/ eval e2
	| Binop ("*",e1,e2) -> eval e1 */ eval e2
	| Binop ("/",e1,e2) -> eval e1 // eval e2
	| Binop (_,e1,e2) -> invalid_arg "eval"
	| Entier n -> Int n;;
let compare_evaluations goal evaluation = compare_num (Int !goal) evaluation
*)

(*
(* A debugging function *)
let rec expr_to_string = function
		Binop (op,e1,e2) -> expr_to_string e1 ^ " " ^ op ^ " " ^ expr_to_string e2
	| Entier n -> string_of_int n;;
*)
let (|>) x f = f x;; (* like a pipe in shell *)
let evalue s = s |> Stream.of_string |> lexer |> read_additive_op |> eval;;

(********* test all possible insertion of operators *********)

type inserted = Nil | Add | Minus | Mult | Div;;

let write operators =
	let result = ref ["9"] in
	for i=8 downto 0 do
		result := (string_of_int i)
			:: (match operators.(i) with Nil->"" |Add->"+" |Minus->" - " |Mult->"*" |Div->"/")
  		:: !result
	done;
String.concat "" !result;;

let goal = ref 0
let nb_sols = ref 0

let test operators =
	let writing = write operators in
	if compare_evaluations !goal (evalue writing) = 0
	then (
(* uncomment to print the solutions *)
(* 		Printf.printf "%s\n%!" writing; *)
		incr nb_sols);;

let operators = Array.make 9 Nil

let rec recurse depth =
if depth = 9
then test operators
else begin
  operators.(depth) <- Nil;   recurse (depth+1);
  operators.(depth) <- Add;   recurse (depth+1);
  operators.(depth) <- Minus; recurse (depth+1);
  operators.(depth) <- Mult;  recurse (depth+1);
  operators.(depth) <- Div;   recurse (depth+1);
end;;


(********* start comptation *********)
let scanf format = Scanf.bscanf Scanf.Scanning.stdib format;;

scanf "%i\n" (fun i -> goal := i);;
recurse 0;;
Printf.printf "%i\n%!" !nb_sols;;

nb_sols := 0;;
scanf "%i\n" (fun i -> goal := i);;
recurse 0;;
Printf.printf "%i\n%!" !nb_sols;;
